from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime


class OnCallEvent(Base):
    """
    SQLAlchemy Model for on-call-events. Meaning, the time a person is on Duty.

    Note: Originally the uid was not added, however, sqladmin does not support composite keys. So this was added to be
    a singular primary key.
    """

    __tablename__ = "on_call_event"
    uid = Column(Integer, primary_key=True, autoincrement=True, unique=True)
    calendar_uid = Column(String(50), ForeignKey("calendar.uid"), nullable=False)
    start_event_utc = Column(UtcDateTime(), nullable=False)
    end_event_utc = Column(UtcDateTime(), nullable=False)
    person_uid = Column(Integer, ForeignKey("person.uid"), nullable=False)
    calendar = relationship("Calendar", back_populates="events")
    person = relationship("Person")

    def __repr__(self):
        return (
            f"OnCallEvent(calendar='{self.calendar_uid}', start_event='{self.start_event_utc}', "
            f"end_event='{self.end_event_utc}', person='{self.person_uid}')"
        )



import os
import typing
from pathlib import Path

from starlette.datastructures import Headers
from starlette.responses import FileResponse, Response
from starlette.staticfiles import NotModifiedResponse, StaticFiles
from starlette.types import Scope

PathLike = typing.Union[str, "os.PathLike[str]"]


class GZIPStaticFiles(StaticFiles):
    """90% is inherited and copied from StaticFiles. We just overwrite this method for GZIP compatibility."""

    def file_response(
        self,
        full_path: PathLike,
        stat_result: os.stat_result,
        scope: Scope,
        status_code: int = 200,
    ) -> Response:
        method = scope["method"]
        request_headers = Headers(scope=scope)

        pathlib_path = Path(full_path)
        if (pathlib_path.parent / f"{pathlib_path.name}.gz").exists():
            gzip_headers = {"content-encoding": "gzip"}
            response = FileResponse(
                str(full_path) + ".gz",
                status_code=status_code,
                stat_result=stat_result,
                method=method,
                headers=gzip_headers,
            )
        else:
            response = FileResponse(full_path, status_code=status_code, stat_result=stat_result, method=method)
        if self.is_not_modified(response.headers, request_headers):
            return NotModifiedResponse(response.headers)
        return response



import logging
import os
from typing import Any, Callable

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import declarative_base, scoped_session
from sqlalchemy.orm import Session as SASession
from sqlalchemy.orm import sessionmaker

log = logging.getLogger(__name__)
SQL_ALCHEMY_CONN: str = os.environ.get("SQL_ALCHEMY_CONNECTION", "")
engine: Engine
Session: Callable[..., SASession]
Base: Any = declarative_base()


def get_engine() -> Engine:
    return engine


def prepare_engine_args():
    """Prepare SQLAlchemy engine args for a postgres connection"""
    return {
        "executemany_mode": "values",
        "executemany_values_page_size": 10000,
        "executemany_batch_page_size": 2000,
        "pool_size": 5,
        "pool_recycle": 1800,
        "pool_pre_ping": True,
        "max_overflow": 10,
        "encoding": "utf-8",
    }


def dispose_orm():
    """Properly close pooled database connections"""
    log.debug("Disposing DB connection pool (PID %s)", os.getpid())
    global engine
    global Session

    if Session:
        Session.remove()
        Session = None
    if engine:
        engine.dispose()
        engine = None


def reconfigure_orm():
    """Properly close database connections and re-configure ORM"""
    dispose_orm()
    configure_orm()


def configure_orm():
    """Configure ORM using SQLAlchemy"""
    log.debug("Setting up DB connection pool (PID %s)", os.getpid())
    global engine
    global Session
    engine_args = prepare_engine_args()
    engine = create_engine(SQL_ALCHEMY_CONN, connect_args={}, **engine_args)
    Session = scoped_session(
        sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=engine,
            expire_on_commit=False,
        )
    )
    create_tables_if_not_present()


def create_tables_if_not_present():
    from duty_board.models.calendar import Calendar
    from duty_board.models.on_call_event import OnCallEvent
    from duty_board.models.person import Person
    from duty_board.models.token import Token

    Person.__table__.create(bind=get_engine(), checkfirst=True)
    Calendar.__table__.create(bind=get_engine(), checkfirst=True)
    OnCallEvent.__table__.create(bind=get_engine(), checkfirst=True)
    Token.__table__.create(bind=get_engine(), checkfirst=True)



from sqlalchemy import Column, Index, String

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime


class Token(Base):
    __table_args__ = (Index("token_last_update_utc", "last_update_utc"),)
    __tablename__ = "token"
    token = Column(String(50), unique=True, primary_key=True)
    username = Column(String(50), unique=True)
    last_update_utc = Column(UtcDateTime(), nullable=False)

    def __repr__(self) -> str:
        return f"Token(token='{self.token}', last_update_utc='{self.last_update_utc}')"



import json
import os
import subprocess
from pathlib import Path
from typing import Any, Dict

from fastapi.openapi.utils import get_openapi

from duty_board.server import app


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x[0].upper() + x[1:] for x in components[1:])


def convert_dict_from_snake_to_camel(given_value: Any) -> Any:
    if isinstance(given_value, (bool, int, float)):
        return given_value
    if isinstance(given_value, list):
        return [
            to_camel_case(item) if isinstance(item, str) else convert_dict_from_snake_to_camel(item)
            for item in given_value
        ]
    if isinstance(given_value, dict):
        return {
            to_camel_case(key): to_camel_case(value)
            if isinstance(value, str)
            else convert_dict_from_snake_to_camel(value)
            for key, value in given_value.items()
        }
    raise ValueError(f"Unexpected value {given_value=}.")


with open("openapi.json", "w") as f:
    openapi_definition: Dict[str, Any] = get_openapi(
        title=app.title,
        version=app.version,
        openapi_version=app.openapi_version,
        description=app.description,
        routes=app.routes,
    )
    import pprint

    pprint.pprint(openapi_definition)
    openapi_camel_case = convert_dict_from_snake_to_camel(openapi_definition)
    json.dump(openapi_camel_case, f)

# consider removing --axios and adding --no-client.
output_folder = "./js/api"
output_file = "api-generated-types.ts"
cmd = f"npx swagger-typescript-api -p ../web_helpers/openapi.json -o {output_folder} -n {output_file} --no-client"
subprocess.run(cmd, cwd=Path.cwd().parent / "www", shell=True)
os.remove(Path.cwd() / "openapi.json")



import importlib
import importlib.machinery
import importlib.util
import logging
import os
import sys
from functools import lru_cache
from inspect import isclass
from typing import List, Optional, Type

from duty_board.exceptions import PluginLoadingException
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.example.example_plugin import ExamplePlugin

logger = logging.getLogger(__name__)


def get_plugin() -> AbstractPlugin:
    loaded_plugin = _get_plugin()
    if loaded_plugin is None:
        logger.warning("Loading the ExamplePlugin as you did not specify a Plugin. This is probably not what you want!")
        return ExamplePlugin()
    return loaded_plugin


@lru_cache
def _get_plugin() -> Optional[AbstractPlugin]:
    if "DUTY_BOARD_PLUGIN_LOCATION" not in os.environ:
        return None

    file_path = os.environ["DUTY_BOARD_PLUGIN_LOCATION"]

    if not os.path.isfile(file_path):
        raise PluginLoadingException(f"{file_path=} is not a file. Please set the plugin location correctly.")
    mod_name, file_ext = os.path.splitext(os.path.split(file_path)[-1])
    if file_ext != ".py":
        raise PluginLoadingException(f"{file_path=} must be a Python file. Please set the plugin correctly.")
    try:
        loader = importlib.machinery.SourceFileLoader(mod_name, file_path)
        spec = importlib.util.spec_from_loader(mod_name, loader)  # type: ignore
        if spec is None:
            raise PluginLoadingException(f"Unable to import {mod_name}.")
        mod = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = mod
        loader.exec_module(mod)
        logger.info("Importing plugin module %s", file_path)
        list_with_plugins: List[Type[AbstractPlugin]] = [
            m for m in mod.__dict__.values() if isclass(m) and issubclass(m, AbstractPlugin) and m != AbstractPlugin
        ]
        if not any(list_with_plugins):
            raise PluginLoadingException(f"{file_path} does not contain any classes that implement the AbstractPlugin.")
        if len(list_with_plugins) > 1:
            raise PluginLoadingException(f"{file_path} contains multiple AbstractPlugin implementations.")
        your_plugin = list_with_plugins[0]
        return your_plugin()
    except PluginLoadingException:
        raise
    except Exception as e:
        logger.exception("Failed to import plugin %s", file_path)
        raise PluginLoadingException(f"Unable to import plugin {file_path}") from e



import asyncio
import os
import secrets
import string
from typing import Optional

from fastapi import FastAPI
from pendulum import DateTime
from sqladmin import Admin, ModelView
from sqladmin.authentication import AuthenticationBackend
from sqladmin.fields import DateTimeField
from sqlalchemy.orm import session as SASession
from starlette.requests import Request

from duty_board.alchemy import settings
from duty_board.alchemy.session import create_session
from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import utc
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.models.token import Token
from duty_board.plugin.abstract_plugin import AbstractPlugin

loop = asyncio.get_event_loop()


class AppBuilderDateTimeAwareSelector(DateTimeField):
    """This is a DateTimePicker"""

    def process_formdata(self, valuelist):
        """This solves the issue with editing a record. Here the timezone would show up. We don't want that!"""
        super().process_formdata(valuelist)
        if self.data is not None:
            self.data = self.data.astimezone(utc)


class PersonAdmin(ModelView, model=Person):  # type: ignore
    column_searchable_list = [Person.username, Person.email, Person.sync]
    column_sortable_list = [Person.uid, Person.username, Person.email, Person.last_update_utc, Person.sync]
    column_list = [Person.uid, Person.username, Person.email, Person.img_filename, Person.last_update_utc, Person.sync]
    form_overrides = dict(last_update_utc=AppBuilderDateTimeAwareSelector)


class CalendarAdmin(ModelView, model=Calendar):  # type: ignore
    column_searchable_list = [Calendar.uid, Calendar.name, Calendar.sync, Calendar.category]
    column_sortable_list = [
        Calendar.uid,
        Calendar.name,
        Calendar.last_update_utc,
        Calendar.sync,
        Calendar.category,
        Calendar.order,
        Calendar.error_msg,
    ]
    column_list = [Calendar.uid, Calendar.name, Calendar.last_update_utc, Calendar.sync]
    form_columns = [  # Adds an order to it :)
        Calendar.uid,
        Calendar.name,
        Calendar.description,
        Calendar.icalendar_url,
        Calendar.event_prefix,
        Calendar.category,
        Calendar.order,
        Calendar.error_msg,
        Calendar.last_update_utc,
        Calendar.sync,
    ]
    form_overrides = dict(last_update_utc=AppBuilderDateTimeAwareSelector)
    form_include_pk = True


class OnCallEventAdmin(ModelView, model=OnCallEvent):  # type: ignore
    column_searchable_list = [
        OnCallEvent.calendar_uid,
        OnCallEvent.start_event_utc,
        OnCallEvent.end_event_utc,
        OnCallEvent.person_uid,
    ]
    column_sortable_list = [OnCallEvent.calendar_uid, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc]
    column_list = [OnCallEvent.calendar, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc, OnCallEvent.person]
    form_columns = [OnCallEvent.calendar, OnCallEvent.start_event_utc, OnCallEvent.end_event_utc, OnCallEvent.person]
    form_overrides = dict(
        start_event_utc=AppBuilderDateTimeAwareSelector, end_event_utc=AppBuilderDateTimeAwareSelector
    )
    form_include_pk = True
    form_ajax_refs = {
        "person": {
            "fields": ("username", "email"),
        }
    }


class TokenAdmin(ModelView, model=Token):  # type: ignore
    can_create = False
    can_edit = False
    can_delete = True
    column_list = [Token.username, Token.token, Token.last_update_utc]


def add_sqladmin(app: FastAPI, plugin: AbstractPlugin) -> Admin:
    # Create the tables if that was not done already
    Base.metadata.create_all(settings.get_engine())

    authentication_backend = MyBackend(plugin=plugin)
    admin = Admin(app=app, engine=settings.get_engine(), authentication_backend=authentication_backend)
    admin.add_view(PersonAdmin)
    admin.add_view(CalendarAdmin)
    admin.add_view(OnCallEventAdmin)
    admin.add_view(TokenAdmin)
    return admin


class MyBackend(AuthenticationBackend):
    def __init__(self, plugin: AbstractPlugin, *args, **kwargs):
        super().__init__(secret_key=os.environ["DUTY_BOARD_SECRET_KEY"])
        self.plugin = plugin

    @staticmethod
    def create_token(username: str) -> str:
        session: SASession
        with create_session() as session:
            session.query(Token).where(Token.username == username).delete()
        with create_session() as session:
            new_token = Token(
                token="".join(secrets.choice(string.ascii_lowercase) for i in range(50)),
                username=username,
                last_update_utc=DateTime.utcnow(),
            )
            session.merge(new_token)
        return new_token.token

    def verify_token(self, provided_token: str) -> bool:
        with create_session() as session:
            token: Optional[Token] = session.query(Token).filter(Token.token == provided_token).first()
            if not token:
                return False
            elif token.last_update_utc < DateTime.utcnow() - self.plugin.admin_session_length:
                session.delete(token)
                return False
            return True

    async def login(self, request: Request) -> bool:
        form = await request.form()
        username = form["username"]
        password = form["password"]
        if not isinstance(username, str) or not isinstance(password, str):
            raise ValueError(
                f"Got interesting types for {username=} and {password=}, {type(username)=}, {type(password)=}"
            )

        if not await self.plugin.admin_login_attempt(username=username, password=password):
            return False

        loop = asyncio.get_event_loop()
        result = await loop.run_in_executor(None, MyBackend.create_token, username)
        request.session.update({"token": result})
        return True

    async def logout(self, request: Request) -> bool:
        request.session.clear()
        return True

    async def authenticate(self, request: Request) -> bool:
        token = request.session.get("token")
        if not await loop.run_in_executor(None, self.verify_token, token):
            request.session.clear()
            return False
        return True



import logging
import subprocess
import sys
from pathlib import Path

import click

from duty_board import worker_loop

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
)
logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command()
def worker():
    logger.info("Starting the worker")
    worker_loop.enter_loop()


@cli.command()
@click.option("--host", default="0.0.0.0", help="The IP address range to listen for.")
@click.option("--port", default="80", type=int, help="The port to listen for.")
@click.option("--reload", is_flag=True, type=bool, help="Auto reload when changes are made to the source repository.")
def webserver(host: str, port: int, reload: bool):
    logger.info("Starting the webserver")
    command = f"uvicorn duty_board.server:app --host {host} --port {port}"
    if reload:
        command += " --reload"

    cwd = str(Path(__file__).absolute())
    with subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1, shell=True, text=True, cwd=cwd
    ) as process:
        if process.stdout is None:
            raise ValueError("Something went wrong without opening the pipeline.")
        for line in iter(process.stdout.readline, ""):
            print(line, end="")
        return_code = process.wait()
        if return_code is not None and return_code != 0:
            raise ChildProcessError(f"Process ended with problems. {return_code=}.")


if __name__ == "__main__":
    cli()



import json
import logging
import os
from functools import wraps
from pathlib import Path
from typing import Any, ClassVar, Dict, Final, List, Literal, Mapping, Optional, Tuple, Union

from ldap3 import Connection, Server, SUBTREE
from ldap3.core import exceptions
from sqlalchemy.orm import Session as SASession

from duty_board.models.person import Person

logger = logging.getLogger(__name__)


def ldap_rebind(func):
    @wraps(func)
    def wrapper(ldap_instance: "LdapBaseClient", *args: Any, **kwargs: Any) -> Any:
        try:
            return func(ldap_instance, *args, **kwargs)
        except (exceptions.LDAPSocketOpenError, exceptions.LDAPSessionTerminatedByServerError):
            ldap_instance.connection.unbind()
            ldap_instance.connection.bind()
            return func(ldap_instance, *args, **kwargs)

    return wrapper


class LdapBaseClient:
    def __init__(
        self,
        url: str,
        domain_components: str,
        user_organizational_unit: str,
        full_quantified_username: str,
        password: str,
        auto_referrals: bool = False,
    ):
        self.url: Final[str] = url
        self.domain_components: Final[str] = domain_components
        self.user_organizational_unit: Final[str] = user_organizational_unit
        self.connection = self._create_connection(full_quantified_username, password, auto_referrals=auto_referrals)

    def get_user(self, user_search_term: str) -> Optional[List[Tuple[str, Mapping[str, List[str]]]]]:
        if "." and "@" in user_search_term:
            search_filter = f"(mail={user_search_term})"
        elif "@" in user_search_term:
            search_filter = f"(uid={user_search_term.split('@')[0]})"
        else:
            search_filter = f"(uid={user_search_term})"

        if self.user_organizational_unit:
            search_base = f"{self.user_organizational_unit},{self.domain_components}"
        else:
            search_base = self.domain_components
        return self._search(search_filter, search_base, "*")

    def _create_connection(self, bind_dn: str, password: str, auto_referrals: bool) -> Connection:
        try:
            server = Server(host=self.url)
            connection = Connection(
                server=server,
                user=bind_dn,
                password=password,
                auto_referrals=auto_referrals,
                raise_exceptions=True,
            )
            connection.bind()
        except exceptions.LDAPInvalidServerError as exc:
            raise exceptions.LDAPException(f"Unable to connect to ldap url: {self.url}.") from exc
        except exceptions.LDAPInvalidCredentialsResult as exc:
            raise exceptions.LDAPException(f"Incorrect credentials for ldap url: {self.url}.") from exc
        except exceptions.LDAPInvalidDNSyntaxResult as exc:
            raise exceptions.LDAPException(f"Invalid bind user: {bind_dn}.") from exc
        if not connection.bound:
            raise exceptions.LDAPException(f"LDAP not bound for url {self.url}")
        return connection

    @ldap_rebind
    def _search(
        self,
        filter_str: str,
        base: str,
        attr_list: Optional[List[str]] = None,
        scope: Literal["BASE", "LEVEL", "SUBTREE"] = SUBTREE,
    ) -> Optional[List[Tuple[str, Mapping[str, List[str]]]]]:
        search_success = self.connection.search(base, filter_str, search_scope=scope, attributes=attr_list)
        if not search_success or not self.connection.response or not self.connection.response[0]:
            return None
        return [(item["dn"], item["attributes"]) for item in self.connection.response]


class LDAPPluginMixin:
    # LDAP settings
    LDAP_URL: ClassVar[str] = "ldaps://some.ldap.server.com:636"
    LDAP_DOMAIN_COMPONENTS: ClassVar[str] = "dc=your,dc=com"
    LDAP_USER_ORGANIZATIONAL_UNIT: ClassVar[str] = "ou=People"
    # This should be the same as you mention in your Plugin.
    LOCATION_TO_STORE_PHOTOS: ClassVar[Path] = Path("/tmp/photos")

    def __init__(self, *args, **kwargs):
        self.client = LdapBaseClient(
            url=self.LDAP_URL,
            domain_components=self.LDAP_DOMAIN_COMPONENTS,
            user_organizational_unit=self.LDAP_USER_ORGANIZATIONAL_UNIT,
            full_quantified_username=os.environ["LDAP_FULL_QUANTIFIED_USERNAME"],
            password=os.environ["LDAP_PASSWORD"],
            auto_referrals=False,
        )
        super(LDAPPluginMixin, self).__init__(*args, **kwargs)

    def _write_image_attribute_to_file(
        self, person: Person, person_attributes: Mapping[str, Union[str, List[str]]]
    ) -> Optional[str]:
        if "jpegPhoto" not in person_attributes:
            return None

        filepath = self.LOCATION_TO_STORE_PHOTOS / f"{person.uid}.jpg"
        with open(filepath, "wb") as file:
            img_as_b64: bytes = person_attributes["jpegPhoto"][0]  # type: ignore
            file.write(img_as_b64)
        return filepath.name

    def _extract_user_info(self, username: str) -> Tuple[str, Mapping[str, Union[str, List[str]]]]:
        result = self.client.get_user(username)
        if not result:
            raise ValueError(f"Could not find user {username}.")
        if len(result) > 1:
            raise ValueError(f"Found multiple people.. {[person[0] for person in result]}.")
        person_ldap_details = result[0]
        dn, attributes = person_ldap_details
        return dn, attributes

    def _get_extra_attributes(
        self, username: str, attributes: Mapping[str, Union[str, List[str]]]
    ) -> Dict[str, Dict[str, str]]:
        """You can use any FontAwesome icons listed here; https://react-icons.github.io/react-icons/icons?name=fa"""
        return {
            "fullName": {
                "information": attributes["cn"][0],
                "icon": "FaUserCircle",
            },
            "location": {
                "information": attributes["l"][0],
                "icon": "FaMapMarkerAlt",
            },
        }

    def sync_person(self, person: Person, session: SASession) -> Person:
        dn, attributes = self._extract_user_info(person.username or person.email)
        person.username = dn.split("=")[1].split(",")[0]  # uid=abc,ou= -> extracts abc
        person.email = attributes["mail"][0]
        person.img_filename = self._write_image_attribute_to_file(person, attributes)
        person.extra_attributes_json = json.dumps(self._get_extra_attributes(person.username, attributes))
        logger.debug(f"Updating references of {person}.")
        return person


# Left this code here, so you can copy-paste this code to your extension and check whether your setup works.
# This requires you to set the `LDAP_FULL_QUANTIFIED_USERNAME` and `LDAP_PASSWORD` environment variables though.
if __name__ == "__main__":
    example_person = Person(ldap="some_username")
    plugin = LDAPPluginMixin()
    from sqlalchemy.orm import create_session

    with create_session() as session:
        plugin.sync_person(example_person, session=session)
        import pprint

        pprint.pprint(example_person)



import json
import random
from datetime import timedelta
from typing import List

from pendulum import DateTime
from sqlalchemy.orm import Session as SASession

from duty_board.alchemy.session import provide_session
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person


@provide_session
def create_calendars(session: SASession) -> None:
    nr_1 = Calendar(
        uid="data_platform_duty",
        name="Data Platform Duty",
        description="If you have any issues with Spark, Airflow or Jupyterhub, contact these guys. "
        "They are available 24/7, however, they work in AMS hours. Only call them if there is an "
        "emergency.",
        category="Big Data",
        order=1,
        last_update_utc=DateTime.utcnow(),
        sync=False,
    )
    nr_2 = Calendar(
        uid="infrastructure_duty",
        name="Infrastructure",
        description="If you have any issues with Firewalls or critical tools such as Yarn, HDFS etc."
        "However, usually Data Platform Duty is your first point of contact for application "
        "issues.",
        category="Infrastructure",
        order=2,
        last_update_utc=DateTime.utcnow(),
        sync=False,
    )
    nr_3 = Calendar(
        uid="machine_learning",
        name="Machine learning",
        description="Do you have any issues with your machine learning tools? Ask these guys.",
        category="Big Data",
        order=3,
        last_update_utc=DateTime.utcnow(),
        sync=False,
    )
    session.merge(nr_1)
    session.merge(nr_2)
    session.merge(nr_3)


@provide_session
def create_persons(session: SASession) -> None:
    nr_1 = Person(
        username="jorrick",
        email="jorricks@gmail.com",
        img_filename="bert.jpeg",
        extra_attributes_json=json.dumps(
            {
                "mattermost": {
                    "information": "Send Mattermost message",
                    "icon": "FaExchangeAlt",
                    "icon_color": "#2980B9",
                    "url": "https://mattermost.com/jorrick",
                },
            }
        ),
        last_update_utc=DateTime.utcnow(),
        sync=False,
    )
    nr_2 = Person(
        username="bart",
        email="bart@gmail.com",
        img_filename="bert.jpeg",
        extra_attributes_json=json.dumps(
            {
                "mattermost": {
                    "information": "Send Mattermost message",
                    "icon": "FaExchangeAlt",
                    "icon_color": "#2980B9",
                    "url": "https://mattermost.com/bart",
                },
            }
        ),
        last_update_utc=DateTime.utcnow(),
        sync=False,
    )
    session.merge(nr_1)
    session.merge(nr_2)


@provide_session
def create_on_call_events(person_uids: List[int], session: SASession) -> None:
    items_to_add = []
    for calendar_uid in ["data_platform_duty", "infrastructure_duty", "machine_learning"]:
        nr_1 = OnCallEvent(
            calendar_uid=calendar_uid,
            start_event_utc=DateTime.utcnow(),
            end_event_utc=DateTime.utcnow() + timedelta(days=1),
            person_uid=random.choice(person_uids),
        )
        nr_2 = OnCallEvent(
            calendar_uid=calendar_uid,
            start_event_utc=DateTime.utcnow() + timedelta(days=1),
            end_event_utc=DateTime.utcnow() + timedelta(days=2),
            person_uid=random.choice(person_uids),
        )
        nr_3 = OnCallEvent(
            calendar_uid=calendar_uid,
            start_event_utc=DateTime.utcnow() + timedelta(days=2),
            end_event_utc=DateTime.utcnow() + timedelta(days=4),
            person_uid=random.choice(person_uids),
        )
        nr_4 = OnCallEvent(
            calendar_uid=calendar_uid,
            start_event_utc=DateTime.utcnow() + timedelta(days=4),
            end_event_utc=DateTime.utcnow() + timedelta(days=5),
            person_uid=random.choice(person_uids),
        )
        items_to_add.extend([nr_1, nr_2, nr_3, nr_4])
    session.bulk_save_objects(items_to_add)


@provide_session
def create_fake_database_rows_if_not_present(session: SASession) -> None:
    if session.query(Calendar).count() < 3:
        create_calendars()
    if session.query(Person).count() < 2:
        create_persons()
    person_uids = [uid[0] for uid in session.query(Person).values(Person.uid)]
    if session.query(OnCallEvent).count() < 5:
        create_on_call_events(person_uids)



from __future__ import annotations

import logging
from typing import List, Optional

import pytz
import requests  # type: ignore
from ical_library import client
from ical_library.ical_components import VCalendar, VEvent
from ical_library.timeline import Timeline
from pendulum import DateTime, Duration
from sqlalchemy.orm import Session as SASession

from duty_board.alchemy.session import create_session
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person

logger = logging.getLogger(__name__)


class ICalPluginMixin:
    @staticmethod
    def _get_icalendar(icalendar_url: str) -> str:
        response = requests.get(icalendar_url, timeout=5.0)
        response.raise_for_status()
        return response.text

    def _get_events_for_upcoming_month(self, icalendar_url: str, event_prefix: str, limit: int = 10) -> List[VEvent]:
        """Gets the calendar events for the upcoming 4 weeks."""
        logger.info(f"Loading {icalendar_url}, this might take some time.")
        now = DateTime.now()
        month_from_now = now + Duration(days=7 * 4)
        calendar_txt: str = self._get_icalendar(icalendar_url=icalendar_url)
        calendar: VCalendar = client.parse_lines_into_calendar(calendar_txt)
        # timeline is ordered by start date.
        timeline: Timeline = calendar.get_limited_timeline(now, month_from_now)
        return [
            event
            for event in timeline.overlapping(now, month_from_now)
            if isinstance(event, VEvent)
            and event.summary is not None
            and event.summary.value is not None
            and event.summary.value.startswith(event_prefix)
        ][:limit]

    def _get_or_create_person(self, value: str) -> int:
        """Returns a Person object if the person is not already in the database."""
        session: SASession
        with create_session() as session:
            if "@" in value:
                result = session.query(Person).filter(Person.email == value).first()
            else:
                result = session.query(Person).filter(Person.username == value).first()
            if result:
                return result.uid
            username = None if "@" in value else value
            email = value if "@" in value else None
            person = Person(
                username=username,
                email=email,
                img_filename=None,
                extra_attributes_json=None,
                last_update_utc=DateTime(1970, 1, 1, 0, 0, 0, tzinfo=pytz.UTC),
                sync=True,
            )
            session.add(person)
        # This extra call is needed to fetch the UID that is automatically created for the Person.
        with create_session() as session:
            result = session.query(Person).filter(Person.username == username).filter(Person.email == email).first()
            if result is None:
                raise ValueError(f"We just added Person {username=} {email=} and we already can't find him anymore..")
            return result.uid

    def _create_on_call_event(self, calendar: Calendar, v_event: VEvent, person_uid: int) -> OnCallEvent:
        return OnCallEvent(
            calendar_uid=calendar.uid,
            start_event_utc=v_event.start,
            end_event_utc=v_event.end,
            person_uid=person_uid,
        )

    def sync_calendar(self, calendar: Calendar, event_prefix: str | None, session: SASession) -> Calendar:
        items_to_insert: List[OnCallEvent] = []
        session.query(OnCallEvent).filter(OnCallEvent.calendar_uid == calendar.uid).delete()
        event: VEvent
        for event in self._get_events_for_upcoming_month(calendar.icalendar_url, event_prefix or ""):
            # First attempt attendee. If that is not set, we look at the title/summary of the event.
            person_unique_identifier: Optional[str] = None
            if event.attendee is not None and any(event.attendee):
                person_unique_identifier = event.attendee[0].value
            if person_unique_identifier is None:
                assert event.summary is not None and event.summary.value is not None  # for Mypy :)
                person_unique_identifier = event.summary.value
                if not person_unique_identifier:
                    raise ValueError(f"{event.summary.value=} should not be None.")
            person_information_str = person_unique_identifier[len(event_prefix or "") :]
            person_uid: int = self._get_or_create_person(person_information_str)
            on_call_event: OnCallEvent = self._create_on_call_event(calendar, event, person_uid=person_uid)
            items_to_insert.append(on_call_event)
        session.bulk_save_objects(items_to_insert)
        return calendar



from __future__ import annotations

from attrs import define, field, validators


@define
class DutyCalendarConfig:
    # A unique identifier that stays the same. When renaming the name object and keeping the same uid, it won't remove
    # all your events & persons in the meantime.
    uid: str = field(validator=[validators.instance_of(str), validators.max_len(50)])  # type: ignore
    # The name shown on the UI.
    name: str = field(validator=[validators.instance_of(str), validators.max_len(200)])  # type: ignore
    # A description about who owns what.
    description: str | None = field(
        validator=validators.optional([validators.instance_of(str), validators.max_len(5000)])  # type: ignore
    )
    # iCalendar URL. This has to be something our request library can fetch.
    icalendar_url: str = field(validator=[validators.instance_of(str), validators.max_len(500)])  # type: ignore
    # The category this duty belongs in (this is what menu item it is shown for)
    category: str | None = field(
        default="default", validator=[validators.instance_of(str), validators.max_len(50)]
    )  # type: ignore
    # Priority of order. The lower the number, the earlier this calendar shows up.
    order: int = field(
        default=99999, validator=[validators.instance_of(int), validators.ge(0), validators.le(9999999)]
    )  # type: ignore
    # Prefix before user LDAP or user email is mentioned. Example prefix; 'duty:' when calendar event; 'duty: thomas'
    event_prefix: str | None = field(
        default=None,
        validator=validators.optional([validators.instance_of(str), validators.max_len(50)]),  # type: ignore
    )  # type: ignore



import datetime

import pendulum as pendulum
from sqlalchemy import DateTime, TypeDecorator

utc = pendulum.tz.timezone("UTC")


class UtcDateTime(TypeDecorator):
    """
    Almost equivalent to :class:`~sqlalchemy.types.DateTime` with
    ``timezone=True`` option, but it differs from that by:

    - Never silently take naive :class:`~datetime.datetime`, instead it
      always raise :exc:`ValueError` unless time zone aware value.
    - :class:`~datetime.datetime` value's :attr:`~datetime.datetime.tzinfo`
      is always converted to UTC.
    - Unlike SQLAlchemy's built-in :class:`~sqlalchemy.types.DateTime`,
      it never return naive :class:`~datetime.datetime`, but time zone
      aware value, even with SQLite or MySQL.
    - Always returns DateTime in UTC

    """

    impl = DateTime(timezone=True)

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            if not isinstance(value, datetime.datetime):
                raise TypeError("expected datetime.datetime, not " + repr(value))
            elif value.tzinfo is None:
                raise ValueError("naive datetime is disallowed")
            return value.astimezone(utc)
        return None

    def process_result_value(self, value, dialect):
        """
        Processes DateTimes from the DB making sure it is always
        returning UTC. Not using timezone.convert_to_utc as that
        converts to configured TIMEZONE while the DB might be
        running with some other setting. We assume UTC datetimes
        in the database.
        """
        if value is not None:
            if value.tzinfo is None:
                value = value.replace(tzinfo=utc)
            else:
                value = value.astimezone(utc)

        return value
